#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#ifndef WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#else
#include <winsock2.h>
#include <ws2tcpip.h>
#endif

#include "socket.h"

static void error_exit(char *message)
{
  fprintf(stderr, "%s", message);
  exit(1);
}

int socket_init()
{
#ifdef WIN32
  WSADATA wsaData;
  if (WSAStartup(MAKEWORD(2, 0), &wsaData) != 0)
    error_exit("WSAStartup() failed.\n");
#endif
  return 0;
}

int socket_open_server(char *hostname, int port)
{
  int s;
  struct hostent *host = NULL;
  struct sockaddr_in addr;
#ifndef WIN32
  int on = 1;
#else
  BOOL on = TRUE;
#endif

  if (hostname) {
    host = gethostbyname(hostname);
    if (host == NULL)
      error_exit("gethostbyname() failed.\n");
  }

  s = socket(AF_INET, SOCK_STREAM, 0);
  if (s < 0)
    error_exit("socket() failed.\n");

#ifndef WIN32
  if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0)
#else
  if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on)) < 0)
#endif
    error_exit("setsockopt() failed.\n");

  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  if (host) {
    memcpy(&addr.sin_addr, host->h_addr, host->h_length);
  } else {
    addr.sin_addr.s_addr = htonl(INADDR_ANY);
  }

  if (bind(s, (struct sockaddr *)&addr, sizeof(addr)) < 0)
    error_exit("bind() failed.\n");

  if (listen(s, 5) < 0)
    error_exit("listen() failed.\n");

  return s;
}

int socket_accept_server(int sockt)
{
  int s;
  struct sockaddr_in addr;
  socklen_t addrlen;
  fd_set fds;
  struct timeval t;

  FD_ZERO(&fds);
  FD_SET(sockt, &fds);
  t.tv_sec  = 0;
  t.tv_usec = 0;
  if (select(sockt + 1, &fds, NULL, NULL, &t) <= 0)
    return -1;
  if (!FD_ISSET(sockt, &fds))
    return -1;

  memset(&addr, 0, sizeof(addr));
  addrlen = sizeof(addr);
  s = accept(sockt, (struct sockaddr *)&addr, &addrlen);
  if (s < 0)
    error_exit("accept() failed.\n");

  return s;
}

int socket_open_dgram_server(char *hostname, int port)
{
  int s;
  struct hostent *host = NULL;
  struct sockaddr_in addr;
#ifndef WIN32
  int on = 1;
#else
  BOOL on = TRUE;
#endif

  if (hostname) {
    host = gethostbyname(hostname);
    if (host == NULL)
      error_exit("gethostbyname() failed.\n");
  }

  s = socket(AF_INET, SOCK_DGRAM, 0);
  if (s < 0)
    error_exit("socket() failed.\n");

#ifndef WIN32
  if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0)
#else
  if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on)) < 0)
#endif
    error_exit("setsockopt() failed.\n");

  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  if (host) {
    memcpy(&addr.sin_addr, host->h_addr, host->h_length);
  } else {
    addr.sin_addr.s_addr = htonl(INADDR_ANY);
  }

  if (bind(s, (struct sockaddr *)&addr, sizeof(addr)) < 0)
    error_exit("bind() failed.\n");

  return s;
}

int socket_open_client(char *hostname, int port)
{
  int s;
  struct hostent *host;
  struct sockaddr_in addr;

  host = gethostbyname(hostname);
  if (host == NULL)
    error_exit("gethostbyname() failed.\n");

  s = socket(AF_INET, SOCK_STREAM, 0);
  if (s < 0)
    error_exit("socket() failed.\n");

  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  memcpy(&addr.sin_addr, host->h_addr, host->h_length);

  if (connect(s, (struct sockaddr *)&addr, sizeof(addr)) < 0)
    error_exit("connect() failed.\n");

  return s;
}

int socket_read(int s, char *buffer, int size, int wait_usec)
{
  fd_set fds;
  struct timeval t;
  int r;

  if (wait_usec >= 0) {
    FD_ZERO(&fds);
    FD_SET(s, &fds);
    t.tv_sec  = wait_usec / 1000000;
    t.tv_usec = wait_usec % 1000000;
    r = select(s + 1, &fds, NULL, NULL, &t);
    if (r <= 0)
      return 0;
    if (!FD_ISSET(s, &fds))
      return 0;
  }

#ifndef WIN32
  r = read(s, buffer, size);
#else
  r = recv(s, buffer, size, 0);
#endif
  if (r < 0)
    return 0;
  return (r == 0) ? -1 : r;
}

int socket_write(int s, char *buffer, int size)
{
#ifndef WIN32
  return write(s, buffer, size);
#else
  return send(s, buffer, size, 0);
#endif
}

static char *skip_space(char *p)
{
  while (*p) {
    if (!isspace(*p))
      break;
    p++;
  }
  return p;
}

#if 0
static char *skip_nospace(char *p)
{
  while (*p) {
    if (isspace(*p))
      break;
    p++;
  }
  return p;
}
#endif

static char *skip_noquote(char *p, int quote)
{
  int escape = 0;
  char *q;

  q = p;
  while (*p) {
    if (!escape) {
      if (*p == '\\') {
	escape = 1;
	*(p++) = '\0';
	continue;
      }
      if (quote) {
	if (*p == quote) break;
      } else {
	if (isspace(*p)) break;
      }
    }
    if ((*p == '\n') || (*p == '\r'))
      break;
    escape = 0;

    if (p != q) *q = *p;
    p++;
    q++;
    if (p != q) *q = '\0';
  }

  return p;
}

char **socket_readline(int *s, int *offset, int *size,
		       int buffer_size, char buffer[], int wait_usec)
{
  static char *argv[16];
  int r, i, quote;
  char *p;

  if (*offset > 0) {
    for (i = 0; i < *size; i++) {
      buffer[i] = buffer[*offset + i];
    }
    *offset = 0;
  }

  if (*s < 0) {
    r = 0;
  } else {
    r = socket_read(*s, buffer + *size, buffer_size - *size, wait_usec);
    if (r < 0) {
#ifndef WIN32
      close(*s);
#else
      closesocket(*s);
#endif
      *s = -1;
      r = 0;
    }
  }

  *size += r;

  for (i = 0; i < *size; i++) {
    if ((buffer[i] == '\n') || (buffer[i] == '\r'))
      break;
  }
  if (i == *size) {
    if ((*size == buffer_size) || (*s < 0))
      *size = 0;
    return NULL;
  }
  buffer[i] = '\0';

  *offset = i + 1;
  *size -= *offset;

  p = buffer;
  i = 0;

  while (1) {
    p = skip_space(p);
    if (*p == '\0')
      break;

    quote = 0;
    if ((*p == '\'') || (*p == '\"')) {
      quote = *p;
      p++;
    }

    argv[i++] = p;
    if (i == sizeof(argv) / sizeof(argv[0]))
      return NULL;

    p = skip_noquote(p, quote);
    if (*p == '\0')
      break;

    *(p++) = '\0';
  }

  argv[i] = NULL;

  return argv;
}
